	processor	6502

;	MMC64 plugin to playback raw unsigned 8-bit samples
;
;	uses code by Levente Hrsfalvi from C=hacking 20/21
;
;	14.489 / 15.040 kHz on PAL/NTSC
;
;	max. sample size is about 40% of card size,
;	13 MB for cards <= 32MB
;
;	card		sample		time
;	size		size
;
;	 32 MB		 13 MB		15 minutes
;	 64 MB		 26 MB		30 minutes
;	128 MB		 53 MB		 1 hour
;	256 MB		106 MB		 2 hours
;	512 MB		212 MB		 4 hours
;	  1 GB		424 MB		 8 hours
;	  2 GB		848 MB		16 hours
;
;
;	to create samples: get sox, http://sox.sourceforge.net/
;
;	sox <infile> -c 1 -r 14489 <outfile.ub>

;	bugs:
;		should end when data ends instead of
;		playing until sector ends
;
;

;	notes:
;		Because of lots of new keyboard code cluster table
;		now overwrites most of init code. This helped keeping
;		cluster table >52 kB.
;
;
;	memory layout
;	$0000..$00ff	zp, mostly intact
;	$0100.,$01ff	stack, zp backup
;	$0200..$03ff	fat/sample buffer
;	$0400..$0427	screen
;	$0428..$2c27	playback code
;	$2c28..$2f??	control code
;	$3000..$ffff	cluster table


	seg.u	zp
	org	$e8

dBase		ds.b	3		; first data sector (from BIOS)
fBase		ds.b	3		; first fat sector (from BIOS)
clustSz		ds.b	1		; sectors per cluster (from BIOS)
fileSz		ds.b	4		; filesize (bytes) (from BIOS)
cluster		ds.b	2		; first cluster (from BIOS)

sec		ds.b	3
fatInMem	ds.b	1

t1		ds.w	1
t2		ds.w	1
t3		ds.w	1

;	only used in player so can reuse cluster scan variables

endPos		= fBase		;.w
endMark		= fBase+2	;.b
cl		= fileSz	;.w
secLeft		= fileSz+2	;.b
zpFF		= fileSz+3	;.b

keys		= t1+1		;.b	!! t1 is used !!
k1		= t2		;.b
k2		= t2+1		;.b
keydelay	= fatInMem

cl_x		= cl-XVAL		; for (zp,x) addressing


fat	=	$0200
buf	=	$0200

XVAL	= $49
YVAL	= $41
VOLREG	= $0f
FILTER	= 0

 mac	play		; sample player by Levente Hrsfalvi
			; this takes 65% of time

	lda	[{1}]	; 4	; get sample
	sta	$d402	; 8	; put it to the sample width reg LO
	stx	$d404	;12	; RESET SID voice
	sty	$d404	;16	; ...then start wave counter
	adc	#$01	;18	; Increase sample value
	sta	$d409	;22	; Fc=0 and stays that way if valid data
	stx	$d40b	;26	; go on for next SID channels
	sty	$d40b	;30
	adc	#$01	;32
	sta	$d410	;36
	stx	$d412	;40
	sty	$d412	;44	;34 bytes
 endm


 mac	get		; store MMC64 byte

	lda	$df10	; 4
	sta	[{1}]	; 8
 endm

 mac	put		; send MMC64 byte

	lda	[{1}]	; 2
	sta	$df10	; 6
 endm

;	various delays

 mac	d3
	bit	$bf	; 3
 endm

 mac	d4
	lda	$bf,x	; 4
 endm

 mac	d5
	asl	t1	; 5
 endm

 mac	d6
	lda	($bf,x)	; 6
 endm

 mac	d7
	pha		; 3
	pla		; 7
 endm

 mac	d8
	lda	($bf,x)	; 6
	nop		; 8
 endm

 mac	d9
	pha		; 3
	pla		; 7
	nop		; 9
 endm

 mac	d10
	lda	($bf,x)	; 6
	lda	$bf,x	;10
 endm

 mac	d18
	lda	($bf,x)	; 6
	lda	($bf,x)	;12
	lda	($bf,x)	;18
 endm

 mac	d21
	lda	#%00000111
	lsr
	bcs	*-1	;21 ( 2+2+2 + 3*5 )
 endm

 mac	d24
	lda	#%00000111
	lsr
	bcs	*-1	;21 ( 2+2+2 + 3*5 )
	bit	$bf	;24
 endm

b1	= buf+1
b2	= buf+2
b3	= buf+3
b4	= buf+4
b170	= buf+170
b171	= buf+171
b172	= buf+172
b173	= buf+173
b174	= buf+174
b175	= buf+175
b176	= buf+176
b177	= buf+177
b178	= buf+178
b179	= buf+179
b180	= buf+180
b181	= buf+181
b182	= buf+182
b183	= buf+183
b184	= buf+184
b185	= buf+185
b186	= buf+186
b187	= buf+187
b188	= buf+188
b189	= buf+189
b190	= buf+190
b191	= buf+191
b192	= buf+192
b193	= buf+193
b194	= buf+194
b195	= buf+195
b196	= buf+196
b197	= buf+197
b198	= buf+198
b199	= buf+199
b256	= buf+256
b511	= buf+511


Playback	= $0428


	seg	code
	org	Playback+170*52-3

	jmp	Start

	subroutine

 if 0			; first 170 samples are played with generated code

;	play first sample

;	we are in read trigger mode, ready to get the first data byte

.loop	play	$df10		;44
	get	b1		;52
	get	b2		;60
	get	b3		;68

s set b1
g set b4
	repeat 170-1
	play	s
	get	g
	get	g+1
	get	g+2
s set s+1
g set g+3
	repend
 endif

;	play	171st sample, decide if cluster continues

	play	b170		;44
	get	b511		;52
	lda	$df10-XVAL,x	;57	; 2nd CRC byte

	dec	secLeft-XVAL,x	;63
	beq	.newc		;65
	jmp	.nexts		;68

.newc				;66
	nop			;68

;	get next cluster-2

	play	b171		;44
	dec	1		;49	; all RAM
	lda	(cl_x,x)	;55	; cluster low
	sta	sec		;58
	inc	1		;63	; I/O back
	inc	cl		;68	; bump pointer (doesn't overflow)

	play	b172		;44
	dec	1		;49
	lda	(cl_x,x)	;55
	sta	sec+1		;58
	inc	1		;63
	inc	cl		;68

	play	b173		;44
	lda	#0		;46	; bump pointer high byte
	sta	sec+2		;49
	cmp	cl-XVAL,x	;53	; Fc=1 if cl=0
	adc	cl+1-XVAL,x	;57
	sta	cl+1-XVAL,x	;61
	d7			;68

	play	b174		;44
	lda	cl-XVAL,x	;48
	cmp	endPos-XVAL,x	;52
	lda	cl+1-XVAL,x	;56
	sbc	endPos+1-XVAL,x	;60	; Fc=1 if end
	lda	clustSz		;63
	ror			;65	; A=clustSz+endFlag<<7, Fc=0
	sta	secLeft		;68

;	shift 1..7 times (512 B .. 32 kB clusters)

	play	b175		;44
	asl	sec		;49
	rol	sec+1		;54
	rol	sec+2		;59	; 512 bytes / cluster
	d9			;68

	play	b176		;44
cl1k	asl	sec		;49
	rol	sec+1		;54
	rol	sec+2		;59	; 1 kB / cluster
	d9			;68

	play	b177		;44
cl2k	asl	sec		;49
	rol	sec+1		;54
	rol	sec+2		;59	; 2 kB / cluster
	d9			;68

	play	b178		;44
cl4k	asl	sec		;49
	rol	sec+1		;54
	rol	sec+2		;59	; 4 kB / cluster
	d9			;68

	play	b179		;44
cl8k	asl	sec		;49
	rol	sec+1		;54
	rol	sec+2		;59	; 8 kB / cluster
	d9			;68

	play	b180		;44
cl16k	asl	sec		;49
	rol	sec+1		;54
	rol	sec+2		;59	; 16 kB / cluster
	d9			;68

	play	b181		;44
cl32k	asl	sec		;49
	rol	sec+1		;54
	rol	sec+2		;59	; 32 kB / cluster
	d9			;68

;	add dBase

	play	b182		;44
	clc			;46
	lda	sec		;49
p_dBl	adc	#0		;51
	sta	sec		;54
	lda	sec+1		;57
p_dBm	adc	#0		;59
	sta	sec+1		;62
	php			;65
	cpx	zpFF		;68	; clc

	play	b183		;44
	plp			;48
	lda	sec+2		;51
p_dBh	adc	#0		;53
	sta	sec+2		;56
	clc			;58	; not really necessary, I hope...
	d7			;65
	jmp	.rdcmd		;68

;	continue with next sector

.nexts	play	b171		;44
	clc			;46
	lda	sec		;49
	adc	#2		;51
	sta	sec		;54
	lda	sec+1		;57
	adc	#0		;59
	sta	sec+1		;62
	php			;65
	cpx	zpFF		;68	; clc

	play	b172		;44
	plp			;48
	lda	sec+2-XVAL,x	;52
	adc	#0		;54
	sta	sec+2-XVAL,x	;58
	clc			;60
	lda	#<b173		;62
	sta	.sloop+1	;66
	nop			;68

;	play samples 173..183 to get to sync with cluster code

.sloop	play	b173		;44
	inc	.sloop+1	;50
	lda	.sloop+1	;54
	eor	#<b184		;56
	d7			;63
	beq	.slx		;65
	jmp	.sloop		;68

.slx				;66
	nop			;68

;	now continue with data read command

.rdcmd	play	b184		;44
	lda	$df11-XVAL,x	;49	; write trigger mode
	and	#~$40		;51
	sta	$df11-XVAL,x	;56
	put	#$ff		;62	; sync
	put	#$51		;68	; read

	play	b185		;44
	lda	sec+2-XVAL,x	;48	; data offset
	sta	$df10		;52
	lda	sec+1-XVAL,x	;56
	sta	$df10		;60
	lda	sec-XVAL,x	;64
	sta	$df10		;68

	play	b186		;44
	put	#0		;50	; <offset
	put	#$ff		;56	; CRC7 + end bit
	sta	$df10-XVAL,x	;61	; sync
	d7			;68

	play	b187		;44
	lda	$df11-XVAL,x	;49	; read trigger mode
	ora	#$40		;51
	sta	$df11-XVAL,x	;56
	lda	$df10-XVAL,x	;61	; fill shift register
	d7			;68

;	do pause if P pressed

	play	b188		;44
	d7			;51
	lda	#$df		;53
	sta	$dc00		;57
	lda	#$02		;59
	and	$dc01		;63
	beq	.pause		;65
	jmp	.fkeys		;68

.pause	lda	#0		; no timing for pause code :)
	sta	$d418
	lda	#$0b
	sta	$d020
	jsr	Pause
	lda	#0
	sta	$d020
	lda	#VOLREG
	sta	$d418
	ldx	#XVAL

;	read function keys, debounce

.fkeys	play	b189		;44
	lda	#$fe		;46
	sta	$dc00-XVAL,x	;51
	lda	$dc01		;55
	cmp	keys		;58
	clc			;60
	sta	keys		;63
	bne	.diff		;65	; restart debounce if different
	jmp	.dofkey		;68

.diff				;66
	nop			;68

	play	b190		;44
	d9			;53
	lda	#2		;55	; reset debounce
	sta	keydelay-XVAL,x	;59
.fback	lda	#<b191		;61	; prepare loop1
	sta	.l1+1		;65
	jmp	.l1		;68	; and jump into it

;	check debounce timer

.dofkey	play	b190		;44
	d5			;49
	dec	keydelay	;54	; check debounce count
	bmi	.f2		;56
	jmp	.fback		;59	; not debounced yet

				;57
.f2	lda	#(1*14489)/512	;59	; repeat delay ~1 second
	sta	keydelay	;62
	d6			;68

	play	b191		;44
	lda	keys		;47	; prepare them for checking
	asl			;49
	sta	k1		;52	; F5 F3 in top bits
	asl			;54
	asl			;56
	sta	k2		;59	; F1 F7 in top bits
	clc			;61
	lda	#<b194		;63	; prepare loop1 for no skip
	sta	.l1+1-XVAL,x	;68

;	check for f5/f3

	play	b192		;44
s10l	lda	#0		;46	; prepare for 10 second skip
	sta	t3		;49
s10h	lda	#0		;51
	sta	t3+1		;54
	bit	k1		;57
	bpl	.f5		;59
	bvc	.f3		;61
	d4			;65
	jmp	.nof53		;68

.f5				;60
	d5			;65	; 10 sec skips have to wait for one
	jmp	.skipf0		;68	; sample to sync with 60 sec skip

.f3				;62
	d3			;65
	jmp	.skipb0		;68

.nof53	play	b193		;44
s60l	lda	#0		;46		prepare for 60 second skip
	sta	t3		;49
s60h	lda	#0		;51
	sta	t3+1		;54
	bit	k2		;57
	bpl	.f1		;59
	bvc	.f7		;61
	d4			;65
	jmp	.l1		;68

.f1				;60
	d5			;65
	jmp	.skipb		;68

.f7				;62
	d3			;65
	jmp	.skipf		;68

;	skip backwards

.skipb0	play	b193		;44
	d24			;68

.skipb	play	b194		;44
	sec			;46
	lda	cl		;49		t3 = cl-t3
	sbc	t3		;52		this never wraps
	sta	t3		;55
	lda	cl+1		;58
	sbc	t3+1		;61
	sta	t3+1		;64
	clc			;66
	nop			;68

;	compare with start, take max(#ctable,t3)

	play	b195		;44
	d7			;51
	lda	t3		;54		t3 >= start ? ok
	cmp	#<ctable	;56
	lda	t3+1		;59
	sbc	#>ctable	;61
	bcc	.start		;63
	clc			;65
	jmp	.tok0		;68

.start				;64
	nop			;66
	nop			;68

	play	b196		;44
	lda	#<ctable	;46
	sta	t3-XVAL,x	;50
	lda	#>ctable	;52
	sta	t3+1-XVAL,x	;56
	d9			;65
	jmp	.tok

;	skip forward

.skipf0	play	b193		;44
	d24			;68

.skipf	play	b194		;44
	clc			;46
	lda	t3		;49		t3 = cl+t3
	adc	cl		;52
	sta	t3		;55
	lda	t3+1		;58
	adc	cl+1-XVAL,x	;62
	sta	t3+1-XVAL,x	;66
	bcs	.skipx		;68

;	compare with end

	play	b195		;44
	lda	t3		;47		t3 >= end ? exit
	cmp	endPos		;50
	lda	t3+1		;53
	sbc	endPos+1	;56
	bcs	.skipx		;58
	d7			;65
	jmp	.tok0		;68

.skipx	jmp	.exit

.tok0	play	b196		;44
	d24			;68

;	ok, force new position after next sector

.tok	play	b197		;44
	lda	t3-XVAL,x	;48
	sta	cl-XVAL,x	;52
	lda	t3+1-XVAL,x	;56
	sta	cl+1-XVAL,x	;60
	lda	#<b199		;62	; prepare loop 1
	sta	.l1+1		;66
	nop			;68

	play	b198		;44
	lda	cl-XVAL,x	;48
	cmp	endPos-XVAL,x	;52
	lda	cl+1-XVAL,x	;56
	sbc	endPos+1-XVAL,x	;60	; Fc=1 if end
	lda	clustSz		;63
	ror			;65	; A=clustSz+endFlag<<7, Fc=0
	sta	secLeft		;68

.l1	play	b189		;44	; play until first page done
	lda	#0		;46	; init loop2
	sta	.l2+1		;50
	d6			;56
	inc	.l1+1		;63
	beq	.l1x		;65
	jmp	.l1		;68

.l1x				;66
	nop			;68

.l2	play	b256		;44	; play until data ready
	inc	.l2+1		;50
	lda	$df10		;54
	eor	#$fe		;56
	beq	.l2x		;58	; exit if data mark found
	d7			;65
	jmp	.l2		;68	; !! no overflow check !!

.l2x				;59
	lda	.l2+1		;63	; init final loop
	sta	.l3+1-XVAL,x	;68

.l3	play	b256		;44
	inc	.l3+1		;50
	beq	.l3x		;52	; check for end
	lda	#$7f		;54
	sta	$dc00-XVAL,x	;59
	lda	$dc01		;63	; check for stop key
	bpl	.exit		;65
	jmp	.l3		;68

.l3x				;53
.mrk	lda	endMark-XVAL,x	;57	; exit if file ended
	eor	secLeft-XVAL,x	;61
	beq	.exit		;63
	nop			;65
	jmp	Playback	;68

;	exit due no more sectors or run/stop
;	we need to read final sector data!

.exit	ldx	#0
.purge	lda	$df10
	lda	$df10
	lda	$df10
	inx
	bne	.purge

	lda	$df11			; write trigger mode
	and	#~$40
	sta	$df11
	rts

;	--------------------------------------------------------

;	ok, here is the entry point

	subroutine
Start
	ldx	#dBase			; save zeropage, copy variables there
.szp	lda	0,x
	pha
	lda	$cf80-dBase,x
	sta	0,x
	inx
	bne	.szp

	jsr	Init
	jsr	BuildCode		; build 8.5 kB of playback code
	jsr	GetSkip			; get 10/60 second skip values
	jsr	PatchPlayer		; fix several self-modifying parts

	jsr	ScanFile

	lda	#0
	sta	t1
	lda	#$ff
	sta	zpFF

.again	jsr	Play

	lda	#$1b
	sta	$d011

	lda	#$7f
	sta	$dc00
.wait	lda	$dc01
	bpl	.x
	and	#$10
	beq	.again
	bne	.wait

.x	ldx	#$1f			; reset SID
	lda	#0
.zero	sta	$d400,x
	dex
	bpl	.zero

	lda	$dd0d			; enable NMI

.rzp	pla				; restore zeropage
	sta	0,x
	dex
	cpx	#dBase
	bcs	.rzp

	lda	#$00			;enable MMC64 BIOS
	sta	$df11
	lda	#$37
	sta	1
	rts

	subroutine
Play
	jsr	WaitFrame
	lda	#0
	sta	$d020
	sta	$d021
	sta	$d011

.replay

;	set pointers

	lda	#<(ctable+2)
	sta	cl
	lda	#>(ctable+2)
	sta	cl+1

;	prepare first sector

	lda	ctable
	sta	sec
	lda	ctable+1
	sta	sec+1
	lda	#0
	sta	sec+2

	lda	clustSz
.1	asl	sec
	rol	sec+1
	rol	sec+2
.2	lsr
	bcc	.1

	ldx	#0
	ldy	#2
	clc
.3	lda	sec,x
	adc	dBase,x
	sta	sec,x
	inx
	dey
	bpl	.3

;	send first read command

	jsr	SendRdCmd

	lda	clustSz
	sta	secLeft

	lda	#VOLREG
	sta	$d418
	ldx	#XVAL		; pre-load X/Y registers
	ldy	#YVAL		; with SID control values

	asl	clustSz		; for player code
	jsr	Playback
	lsr	clustSz		; undo

	lda	#$fd		; check for left shift
	sta	$dc00
	bit	$dc01
	bpl	.replay

.x	lda	#0
	sta	$d418
	rts

	subroutine
SendRdCmd
	lda	$df11		; write trigger mode
	and	#~$40
	sta	$df11
	
	sty	$df10		; read command
	put	#$51
	put	sec+2
	put	sec+1
	put	sec+0
	put	#$00
	sty	$df10

	lda	$df11		; read trigger mode
	ora	#$40
	sta	$df11
.wait	lda	$df10		; wait until data ready
	cmp	#$fe
	bne	.wait
	rts


	subroutine
Pause
.p1	jsr	DebounceKbd
	and	#2
	beq	.p1

.p2	jsr	DebounceKbd
	and	#2
	bne	.p2

.p3	jsr	DebounceKbd
	and	#2
	beq	.p3

	rts

	subroutine
DebounceKbd
.1	lda	$dc01
	ldx	#2			; 2 frames

.2	jsr	WaitFrame
	cmp	$dc01			; restart if changed
	bne	.1

	dex				; wait more if not done yet
	bne	.2

	rts

WaitFrame
	bit	$d011			; wait one frame
	bpl	*-3
	bit	$d011
	bmi	*-3
	rts


;	--------------------------------------------------------


;	we can put cluster table here as everything
;	below is unused after init or copied elsewhere

	align	2
ctable

_ScanFile
	rorg	$0100

;	find sector offsets

	subroutine
ScanFile

	lda	clustSz			; size into pages
	asl
	sta	t2

	lda	cluster+1		; make sure we read FAT on entry
	eor	#$ff
	sta	fatInMem

	lda	#<ctable
	sta	t1
	lda	#>ctable
	sta	t1+1

;	copy offsets to table until file end of mem filled

.1	dec	$01

	sec
	ldy	#0
	lda	cluster
	sbc	#2
	sta	(t1),y			; lo
	iny
	lda	cluster+1
	sbc	#0
	sta	(t1),y			; hi

	inc	$01

	inc	t1
	inc	t1
	bne	.2
	inc	t1+1

.2	sec				; decrement size
	lda	fileSz+1
	sbc	t2
	sta	fileSz+1
	lda	fileSz+2
	sbc	#0
	sta	fileSz+2
	lda	fileSz+3
	sbc	#0
	sta	fileSz+3
	bcc	.end			; exit if file ended
	ora	fileSz+2
	ora	fileSz+1
	ora	fileSz
	beq	.end			; even if it did it just so and so

;	see if memory is full

	lda	t1+1
	beq	.full

;	find next cluster

	jsr	NextCluster
	jmp	.1

;	file ended, calculate end mark = $80 + unused sectors

.end	clc
	lda	fileSz+1
	adc	t2
	sta	fileSz+1

	lda	fileSz
	bne	.e1
	dec	fileSz+1
.e1	lsr	fileSz+1

	sec
	lda	clustSz
	sbc	fileSz+1
	clc
	adc	#$7f

	ldx	t1
	ldy	t1+1
	bne	.e3

;	if memory full we just set end mark

.full	lda	#$80
	ldx	#$fe			; compare doesn't work with $0000
	ldy	#$ff

.e3	sta	endMark
	stx	endPos
	sty	endPos+1
	rts

	subroutine
NextCluster				; fbase=0040

	lda	cluster+1
	cmp	fatInMem
	beq	.3

	sta	fatInMem
	clc
	ldx	#0
	ldy	#2
.1	adc	fBase,x
	sta	sec,x
	lda	#0
	inx
	dey
	bpl	.1

	asl	sec
	rol	sec+1
	rol	sec+2

	jsr	SendRdCmd

	iny
.2	lda	$df10
	sta	fat,y
	lda	$df10
	sta	fat+256,y
	iny
	bne	.2

	lda	$df10			; 2nd CRC byte
	lda	$df11			; write trigger mode
	and	#~$40
	sta	$df11

.3	ldy	cluster
	lda	fat,y
	sta	cluster
	lda	fat+256,y
	sta	cluster+1
	rts

	rend
_ScanFile_e


;	--------------------------------------------------------

	subroutine
Init
	sei
	lda	#$35
	sta	1

	lda	#$0c			; disable MMC64 BIOS and external cart
	sta	$df11

	lda	#<.rti			; disable NMI
	sta	$fffa
	lda	#>.rti
	sta	$fffb
	lda	#$81
	sta	$dd0d			; CIA2 TA IR enable
	ldx	#0			; IRQ in one cycle
	stx	$dd05
	inx
	stx	$dd04
	lda	#$dd
	sta	$dd0e			; cause it now

	lda	#0
	sta	$d020			; all black to hide code
	sta	$d021
	tax
.1	sta	$d800,x
	sta	$d900,x
	sta	$da00,x
	sta	$db00,x
	inx
	bne	.1

	ldx	#39			; text to screen
.2	lda	text,x
	sta	$0400,x
	lda	#15
	sta	$d800,x
	dex
	bpl	.2

.3	inx				; cluster scan into stack
	lda	_ScanFile,x
	sta	$0100,x
	cpx	#_ScanFile_e-_ScanFile
	bne	.3

;	init SID

	lda	#$00			;Set SID freq registers
	sta	$d418
	sta	$d400
	sta	$d407
	sta	$d40e
	lda	#$40			;$4000 --> the wave accu is
	sta	$d401			;incremented by $004 each CPU clock
	sta	$d408			;cycles
	sta	$d40f
	lda	#$f0			;maximum sustain level for all chns
	sta	$d406
	sta	$d40d
	sta	$d414

 if 	FILTER=1
	lda	#7
	sta	$d415
	lda	#$a0
	sta	$d416
	lda	#$07
	sta	$d417
 endif 

	rts

.rti	rti
text	dc.b	" SPACE TO PLAY AGAIN, RUN/STOP TO EXIT  "


	subroutine
BuildCode

	ldx	#<Playback		; code goes here
	ldy	#>Playback
	stx	t1
	sty	t1+1
	ldx	#<buf			; code reads samples
	ldy	#>buf			; from here
	stx	t2
	sty	t2+1
	ldx	#<(buf+1)		; and stores data here
	stx	t3
	sty	t3+1

.0	ldy	#0
.1	lda	.code,y
	beq	.4			; fill in blanks
.2	sta	(t1),y
	iny
	cpy	#.code_-.code
	bcc	.1

	clc
	tya
	adc	t1
	sta	t1
	bcc	.3
	inc	t1+1

.3	ldy	t3
	iny
	bne	.0

	ldx	#$10			; fix first sample location
	ldy	#$df
	stx	Playback+1
	sty	Playback+2
	rts

.4	cpy	#2
	bcs	.5
	lda	t2			; 1 = <sample
	bcc	.2
.5	bne	.6
	lda	t2+1			; 2 = >sample
	inc	t2
	bne	.2
	inc	t2+1
	bne	.2

.6	tya				; even = <store
	lsr				; odd = >store
	bcs	.7
	lda	t3
	bcc	.2
.7	lda	t3+1
	inc	t3
	bne	.2
	inc	t3+1
	bne	.2

.code	lda.w	$0000	; 4	; get sample
	sta	$d402	; 8	; put it to the sample width reg LO
	stx	$d404	;12	; RESET SID voice
	sty	$d404	;16	; ...then start wave counter
	adc	#$01	;18	; Increase sample value
	sta	$d409	;22	; Fc=0 and stays that way if valid data
	stx	$d40b	;26	; go on for next SID channels
	sty	$d40b	;30
	adc	#$01	;32
	sta	$d410	;36
	stx	$d412	;40
	sty	$d412	;44	; 34 bytes

	lda	$df10	;48
	sta.w	$0000	;52
	lda	$df10	;56
	sta.w	$0000	;60
	lda	$df10	;64
	sta.w	$0000	;68	; 52 bytes
.code_	


;	get skip values for PAL/NTSC and this card

	subroutine
GetSkip

	lda	#0
.1	tax
.2	lda	$d012			; remember last >0 $d012 value
	bne	.1
	lda	$d011			; go back if it we are at line 256
	bmi	.2
	txa				; go back if we happened to come
	beq	.2			; here during line 0

	cmp	#<(262+312)/3		; halfway between PAL/NTSC
	ldx	#PALskip-PALskip
	bcs	.3			; ok, it was PAL
	ldx	#NTSCskip-PALskip

;	now get correct timing for this card size

.3	lda	clustSz

.4	lsr
	bcs	.5
	inx
	inx
	inx
	inx
	bne	.4

;	move cluster counts into code, *2 for word table

.5	lda	PALskip,x
	asl
	sta	s10l+1
	lda	PALskip+1,x
	rol
	sta	s10h+1

	lda	PALskip+2,x
	asl
	sta	s60l+1
	lda	PALskip+3,x
	rol
	sta	s60h+1

	rts


PALskip
	dc.w	14489*10/512,  14489*60/512
	dc.w	14489*10/1024, 14489*60/1024
	dc.w	14489*10/2048, 14489*60/2048
	dc.w	14489*10/4096, 14489*60/4096
	dc.w	14489*10/8192, 14489*60/8192
	dc.w	14489*10/16384,14489*60/16384
	dc.w	14489*10/32768,14489*60/32768

NTSCskip
	dc.w	15040*10/512,  15040*60/512
	dc.w	15040*10/1024, 15040*60/1024
	dc.w	15040*10/2048, 15040*60/2048
	dc.w	15040*10/4096, 15040*60/4096
	dc.w	15040*10/8192, 15040*60/8192
	dc.w	15040*10/16384,15040*60/16384
	dc.w	15040*10/32768,15040*60/32768


	subroutine
PatchPlayer

	asl	dBase			; dBase into byte offset / 256
	rol	dBase+1
	rol	dBase+2

	lda	dBase			; copy it into player code
	sta	p_dBl+1
	lda	dBase+1
	sta	p_dBm+1
	lda	dBase+2
	sta	p_dBh+1

	ldx	#t1			; unneeded shifts in cluster size
	lda	clustSz			; shift t1 instead of sec

	asl
	bmi	.1			; 32 kB cluster
	stx	cl32k+1
	stx	cl32k+3
	stx	cl32k+5

	asl
	bmi	.1			; 16 kB cluster
	stx	cl16k+1
	stx	cl16k+3
	stx	cl16k+5

	asl
	bmi	.1			; 8 kB cluster
	stx	cl8k+1
	stx	cl8k+3
	stx	cl8k+5

	asl
	bmi	.1			; 4 kB cluster
	stx	cl4k+1
	stx	cl4k+3
	stx	cl4k+5

	asl
	bmi	.1			; 2 kB cluster
	stx	cl2k+1
	stx	cl2k+3
	stx	cl2k+5

	asl
	bmi	.1			; 1 kB cluster
	stx	cl1k+1
	stx	cl1k+3
	stx	cl1k+5
	
.1	rts
